// ==========================================================================
//                                  FMIndex
// ==========================================================================
// Copyright (c) 2006-2012, Knut Reinert, FU Berlin
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of Knut Reinert or the FU Berlin nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ==========================================================================
// Author: Jochen Singer <jochen.singer@fu-berlin.de>
// ==========================================================================

#ifndef INDEX_FM_LF_TABLE_H_
#define INDEX_FM_LF_TABLE_H_

namespace seqan {

// ==========================================================================
// LfTable is an object storing all necessary information for the LF-mapping.
// To be more precise, the occurrence-table data structure as well as the
// prefix-sum table are stored.
// ==========================================================================

template <typename TText, typename TSpec>
class WaveletTree;

template <typename TSpec>
struct FmiDollarSubstituted;

struct FibreOccTable_;
struct FibrePrefixSumTable_;
struct FibreDollarPosition_;


typedef Tag<FibreOccTable_> const FibreOccTable;
typedef Tag<FibrePrefixSumTable_> const FMTablePrefixSumTable;
typedef Tag<FibreDollarPosition_> const FibreDollarPosition;

template <typename TOccTable, typename TPrefixSumTable>
struct LfTable;

// ==========================================================================
//Metafunctions
// ==========================================================================

template <typename TOccTable, typename TPrefixSumTable>
struct Fibre<LfTable<TOccTable, TPrefixSumTable>, FibreOccTable>
{
    typedef TOccTable Type;
};

template <typename TOccTable, typename TPrefixSumTable>
struct Fibre<LfTable<TOccTable, TPrefixSumTable> const, FibreOccTable>
{
    typedef TOccTable const Type;
};

// ==========================================================================
template <typename TOccTable, typename TPrefixSumTable>
struct Fibre<LfTable<TOccTable, TPrefixSumTable>, FMTablePrefixSumTable>
{
    typedef TPrefixSumTable Type;
};

template <typename TOccTable, typename TPrefixSumTable>
struct Fibre<LfTable<TOccTable, TPrefixSumTable> const, FMTablePrefixSumTable>
{
    typedef TPrefixSumTable const Type;
};

// ==========================================================================
template <typename TOccTable, typename TPrefixSumTable>
struct Reference<LfTable<TOccTable, TPrefixSumTable> >
{
    typedef TPrefixSumTable & Type;
};

// ==========================================================================
// Classes
// ==========================================================================

/**
.Class.LfTable:
..cat:Index
..summary:LfTable is an object storing all necessary information for the LF-mapping.
..signature:LfTable<TOccTable, TPrefixSumTable>
..param.TOccTable:The occurrence table data structure.
...type.class:WaveletTree
..param.TPrefixSumTable:The specialisation tag.
...default:String
..include:seqan/Index.h
*/
template <typename TOccTable, typename TPrefixSumTable>
struct LfTable
{
    TOccTable occTable;
    TPrefixSumTable prefixSumTable;

    LfTable() :
        occTable(),
        prefixSumTable()
    {}

    LfTable(TOccTable const & occTable, TPrefixSumTable const & prefixSumTable) :
        occTable(occTable),
        prefixSumTable(prefixSumTable)
    {}

    inline LfTable & operator=(LfTable const & other)
    {
        occTable = other.occTable;
        prefixSumTable = other.prefixSumTable;
        return *this;
    }

    inline bool operator==(const LfTable & b) const
    {
        return occTable == b.occTable &&
               prefixSumTable == b.prefixSumTable;
    }

};

// ==========================================================================
// Functions
// ==========================================================================

///.Function.clear.param.object.type:Class.LfTable
///.Function.clear.class:Class.LfTable
template <typename TOccTable, typename TPrefixSumTable>
inline void clear(LfTable<TOccTable, TPrefixSumTable> & lfTable)
{
    clear(lfTable.occTable);
    clear(lfTable.prefixSumTable);
}

// ==========================================================================
///.Function.empty.param.object.type:Class.LfTable
///.Function.empty.class:Class.LfTable
template <typename TOccTable, typename TPrefixSumTable>
inline bool empty(LfTable<TOccTable, TPrefixSumTable> & lfTable)
{
    return empty(lfTable.occTable)
           && empty(lfTable.prefixSumTable);
}

// ==========================================================================
/**
.Function.createLfTable
..summary:Creates the LF table
..signature:createLfTable(lfTable, text)
..param.lfTable:The LF table to be constructed.
..type:Class:LfTable.
..param.text:The underlying text.
...type.class:String
..include:seqan/index.h
*/
template <typename TPrefixSumTable, typename TText, typename TDollarSpec, typename TText2>
inline bool createLfTable(LfTable<WaveletTree<TText, FmiDollarSubstituted<TDollarSpec> >, TPrefixSumTable> & lfTable, TText2 const text)
{
    typedef typename SAValue<TText2>::Type   TSAValue;
    typedef typename Value<WaveletTree<TText, FmiDollarSubstituted<TDollarSpec> > >::Type     TAlphabet;
    typedef typename Fibre<WaveletTree<TText, FmiDollarSubstituted<TDollarSpec> >, FibreDollarPosition>::Type TDollarPos;

    String<TSAValue> sa;
    resize(sa, length(text));
    createSuffixArray(sa, text, Skew7());

    createPrefixSumTable(lfTable.prefixSumTable, text);

    TAlphabet dollarSub;
    determineDollarSubstitute_(lfTable.prefixSumTable, dollarSub);

    String<TAlphabet> bwt;
    TDollarPos dollarPos = 0;
    resize(bwt, computeBwtLength_(text));
    createBwTable_(bwt, dollarPos, text, sa, dollarSub);

    clear(sa);

    createOccurrenceTable(lfTable, bwt, dollarSub, dollarPos);
    clear(bwt);

    insertDollar_(lfTable.prefixSumTable, countSequences(text));

    //here comes the dollar modification
    //addDollarNode(index.lfTable.occTable, dollarSub, dollarPos);
    return true;
}

// ==========================================================================
/**
.Function.getFibre
..param.container:
...type:Class.LfTable
*/
template <typename TOccTable, typename TPrefixSumTable>
inline typename Fibre<LfTable<TOccTable, TPrefixSumTable>, FMTablePrefixSumTable>::Type &
getFibre(LfTable<TOccTable, TPrefixSumTable>&lfTable, FMTablePrefixSumTable)
{
    return lfTable.prefixSumTable;
}

template <typename TOccTable, typename TPrefixSumTable>
inline typename Fibre<LfTable<TOccTable, TPrefixSumTable>, FMTablePrefixSumTable>::Type const &
getFibre(LfTable<TOccTable, TPrefixSumTable> const & lfTable, FMTablePrefixSumTable)
{
    return lfTable.prefixSumTable;
}

template <typename TOccTable, typename TPrefixSumTable>
inline typename Fibre<LfTable<TOccTable, TPrefixSumTable>, FibreOccTable>::Type &
getFibre(LfTable<TOccTable, TPrefixSumTable>&lfTable, FibreOccTable)
{
    return lfTable.occTable;
}

template <typename TOccTable, typename TPrefixSumTable>
inline typename Fibre<LfTable<TOccTable, TPrefixSumTable>, FibreOccTable>::Type const &
getFibre(LfTable<TOccTable, TPrefixSumTable> const & lfTable, FibreOccTable )
{
    return lfTable.occTable;
}

// ==========================================================================
/**
.Function.lfMapping:
..summary:Returns the position of the character L[c] in F.
..cat:Index
..signature:lfMapping(lfTable, pos)
..param.lfTable:The @Class.LfTable@ holding the occurrence and prefix sum table.
...type:Class.LfTable
..param.pos:The position in L
..returns:Returns the position of the character L[c] in F. The returned position is of the same type as pos.
..include:seqan/index.h
*/
template <typename TLfTable, typename TPos>
inline TPos lfMapping(TLfTable & lfTable,
                      TPos pos)
{
    typedef typename Fibre<TLfTable, FibreOccTable>::Type TOccTable;
    typedef typename Value<TOccTable>::Type TChar;
    TChar c = getCharacter(lfTable.occTable, pos);
    return getOccurrences(lfTable.occTable, c, pos) + getPrefixSum(lfTable.prefixSumTable, getCharacterPosition(lfTable.prefixSumTable, c)) - 1;
}

// ==========================================================================
/**
.Function.open
..param.string:
...type:Class.LfTable
*/
template <typename TOccTable, typename TPrefixSumTable>
inline bool open(
    LfTable<TOccTable, TPrefixSumTable> & lfTable,
    const char * fileName,
    int openMode)
{
    String<char> name;
    name = fileName; 
    if (!open(getFibre(lfTable, FibreOccTable()), toCString(name), openMode))
    {
        return false;
    }
    name = fileName;    open(getFibre(lfTable, FMTablePrefixSumTable()), toCString(name), openMode);
    return true;

}

template <typename TOccTable, typename TPrefixSumTable>
inline bool open(
    LfTable<TOccTable, TPrefixSumTable> & lfTable,
    const char * fileName)
{
    return open(lfTable, fileName, DefaultOpenMode<LfTable<TOccTable, TPrefixSumTable> >::VALUE);
}

// ==========================================================================
/**
.Function.save
..param.string:
...type:Class.LfTable
*/
template <typename TOccTable, typename TPrefixSumTable>
inline bool save(
    LfTable<TOccTable, TPrefixSumTable> const & lfTable,
    const char * fileName,
    int openMode)
{
    String<char> name;
    name = fileName;
    if (!save(getFibre(lfTable, FibreOccTable()), toCString(name), openMode))
    {
        return false;
    }
    name = fileName;    save(getFibre(lfTable, FMTablePrefixSumTable()), toCString(name), openMode);
    return true;
}

template <typename TOccTable, typename TPrefixSumTable>
inline bool save(
    LfTable<TOccTable, TPrefixSumTable> const & lfTable,
    const char * fileName)
{
    return save(lfTable, fileName, DefaultOpenMode<LfTable<TOccTable, TPrefixSumTable> >::VALUE);
}

}
#endif // INDEX_FM_LF_TABLE_H_
